/*
 * $Id: bcp.c,v 1.2 2001/01/17 19:43:18 mpeppler Exp $
 *
 * bcp.c     - Perform bcp operation from one server to another.
 *
 * Copyright (C) 2001 by Michael Peppler
 *
 * Based on:
 * cmd_bcp.c - User command to bcp result set to another server.
 *
 * Copyright (C) 1995, 1996 by Scott C. Gray
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * You may contact the author :
 *   e-mail:  mpeppler@peppler.org
 *
 */

#include <gtk/gtk.h>

#include <stdio.h>
#include <ctpublic.h>
#include <bkpublic.h>
#include "bcptool.h"
#include "bcp.h"

/*
 * The following macro is used to convert a start time and end
 * time into total elapsed number of seconds, to the decimal
 * place.
 */
#define ELAPSED_SEC(tv_start,tv_end) \
   (((double)(tv_end.tv_sec - tv_start.tv_sec)) + \
      ((double)(tv_end.tv_usec - tv_start.tv_usec)) / ((double)1000000.0))

/*
 * sg_interrupted:  This variable is set by the bcp_signal() signal
 *      handler and is to be polled periodically to determine if 
 *      a signal has been recieved.
 */
static int sg_interrupted  = FALSE;

/*
 * sg_bcp_connection: This variable is used by the signal handler
 *      to determine which connection needs to be canceled.
 */
static CS_CONNECTION *sg_bcp_connection = NULL;
static CS_CONNECTION *sg_dbh_connection = NULL;

#if defined DEBUG
static CS_CONTEXT    *sg_context;
static void
dump_row( bcp_data_t *d);
#endif

/*
 * sg_table: This is the TABLE information for the source & target
 *      table for the request. We need this (local) global variable
 *      to handle IDENTITY columns correctly.
 */
static TABLE *sg_table;


/*-- Prototypes --*/
static void        bcp_signal ( int, void* );
static bcp_data_t* bcp_data_bind ( CS_COMMAND*, CS_INT );
static CS_INT      bcp_data_xfer ( bcp_data_t*, CS_COMMAND*, CS_BLKDESC* );
static void        bcp_data_destroy ( bcp_data_t* );
static char *      build_sql (TABLE *table);
static void        get_target_info (DBH *bcp, TABLE *table);

#if 0
static CS_RETCODE bcp_server_cb ( CS_CONTEXT*, CS_CONNECTION*, CS_SERVERMSG* );
static CS_RETCODE bcp_client_cb ( CS_CONTEXT*, CS_CONNECTION*, CS_CLIENTMSG* );
#endif


int
cmd_bcp (DBH *dbh, DBH *bcp, TABLE *table, char *bcp_table)
{
    char             *cmd_sql;       /* SQL command to send to server */
    int               rows_in_batch; /* Rows processed in batch */
    int               total_rows;    /* Total rows processing */
    struct timeval    tv_start;      /* Time we started processing */
    struct timeval    tv_end;        /* Time we finished processing */
    double            secs;          /* Seconds spend transferring data */
    int               nerrors;       /* Number of errors encountered */
    CS_INT            result_type;   /* Result type coming from server */
    CS_INT            return_code;   /* Return code from ct_results() */
    CS_INT            nrows;         /* Rows transferred */
    CS_INT            con_status;
    CS_INT            value;
    
    /*
     * The following variables need to be initialized to check if
     * they need cleaning up before we exit this procedure.
     */
    CS_COMMAND     *bcp_cmd    = NULL;  /* Command sent to server */
    CS_BLKDESC     *bcp_desc   = NULL;  /* Block descriptor */
    bcp_data_t     *bcp_dat    = NULL;  /* Bind data */
    int               maxerrors     = 10;   /* Errors before abort */
    int               batchsize     = table->batchsize;
//    int               have_error    = FALSE;
    CS_BOOL           have_identity = CS_FALSE;
    
    /*
     * Now, install our signal handlers.  At this point, all code should
     * perform a goto {return_fail or return_interrupt} to return an error
     * condition.  This is important in order to clean up the signal
     * handlers and various other data structures created.
     */
    sg_interrupted    = FALSE;
    sg_bcp_connection = bcp->connection;
    sg_dbh_connection = dbh->connection;

#if defined DEBUG
    sg_context        = dbh->context;
#endif

    /* FIXME */
//    sig_install( SIGINT, bcp_signal, (void*)NULL, 0 );
//    sig_install( SIGPIPE, bcp_signal, (void*)NULL, 0 );

    get_target_info(bcp, table);
    sg_table = table;

    /* XXX: FIXME */
    cmd_sql = build_sql(table);
    
    /*-- Initialize timing --*/
    gettimeofday( &tv_start, NULL );
    total_rows = 0;

    /* Set options on query connections */
    if (table->rowcount) {
	value = table->rowcount;
	if (ct_options(dbh->connection, 
		       CS_SET,
		       CS_OPT_ROWCOUNT,
		       &value,
		       CS_UNUSED,
		       NULL) != CS_SUCCEED) 
	{
	    g_printerr("bcp: Unable to set ROWCOUNT\n");
	    goto return_fail;
	}
    }

#if 0
	/*-- Client callback --*/
	if (ct_callback( (CS_CONTEXT*)NULL,         /* Context */
	                 bcp->connection,           /* Connection */
	                 CS_SET,                    /* Action */
	                 CS_CLIENTMSG_CB,           /* Type */
	                 (CS_VOID*)bcp_client_cb    /* Callback Pointer */
	               ) != CS_SUCCEED)
		goto return_fail;

	/*-- Server callback --*/
	if (ct_callback( (CS_CONTEXT*)NULL,         /* Context */
	                 bcp->connection,           /* Connection */
	                 CS_SET,                    /* Action */
	                 CS_SERVERMSG_CB,           /* Type */
	                 (CS_VOID*)bcp_server_cb    /* Callback Pointer */
	               ) != CS_SUCCEED)
		goto return_fail;             
#endif

    
    /*-- Allocate a new command structure --*/
    if (ct_cmd_alloc( dbh->connection, &bcp_cmd ) != CS_SUCCEED)
    {
	g_printerr( "bcp: Unable to allocate new command\n" );
	goto return_fail;
    }
    
    /*-- Initialize the command --*/
    if (ct_command( bcp_cmd,                /* Command */
		    CS_LANG_CMD,            /* Type */
		    (CS_VOID*)cmd_sql,      /* Buffer */
		    CS_NULLTERM,            /* Buffer Length */
		    CS_UNUSED ) != CS_SUCCEED)
    {
	g_printerr( "bcp: Unable to initialize command structure\n" );
	goto return_fail;
    }
    
    /*-- Send command to server --*/
    if (ct_send( bcp_cmd ) != CS_SUCCEED)
    {
	g_printerr( "bcp: Unable to send command to SQL Server\n" );
	goto return_fail;
    }
    
    if (sg_interrupted)
	goto return_interrupt;

    /* Set target table options */
    if (table->trunc_target) {
	char buff[100];
	RESULT *ret;
	sprintf(buff, "truncate table %s", bcp_table);
	/* XXX - check ret code! */
	ret = db_exec_sql(bcp, buff);
	db_free_result(ret);
    }
    
    /*-- Allocate a block descriptor --*/
    if (blk_alloc( bcp->connection, BLK_VERSION, &bcp_desc ) != CS_SUCCEED)
    {
	g_printerr( "bcp: Unable to allocate bulk descriptor\n" );
	goto return_fail;
    }
	
    /*
     * Configure whether or not this connection is to contain the
     * value for the identity column in a result set.  We default
     * this to true.
     */
    if (table->id_insert) 
	have_identity = CS_TRUE;

    if (blk_props( bcp_desc,                    /* Descriptor */
		   CS_SET,                      /* Action */
		   BLK_IDENTITY,                /* Property */
		   (CS_VOID*)&have_identity,    /* Buffer */  
		   CS_UNUSED,                   /* Buffer Length*/  
		   (CS_INT*)NULL                /* Output Length */
	    ) == CS_FAIL)
    {
	g_printerr( "bcp: Unable to set BLK_IDENTITY option to %s\n",
		    have_identity == CS_TRUE ? "CS_TRUE" : "CS_FALSE" );
	goto return_fail;
    }

    if (blk_init( bcp_desc, CS_BLK_IN, bcp_table, 
		  strlen(bcp_table) ) != CS_SUCCEED)
    {
	g_printerr( "bcp: Unable to initialize bulk copy on table '%s'\n",
		     bcp_table );
	goto return_fail;
    }

    g_printerr("\nStarting copy...\n" );
	
    /*
     * Allrightythen.  We have already sent the command to retrieve
     * data through g_dbproc, and have succesfully created bcp_dbproc
     * in order to jam the data into another database, so lets start
     * processing results.
     */
    rows_in_batch = 0;
    nerrors  = 0;

    while ((return_code = ct_results( bcp_cmd, &result_type )) != CS_END_RESULTS)
    {
	if (sg_interrupted)
	    goto return_interrupt;
		
	if (return_code != CS_SUCCEED)
	    break;

	switch (result_type)
	{
	  case CS_ROW_RESULT:

	    /*-- Destroy old data --*/
	    if (bcp_dat != NULL)
	    {
		bcp_data_destroy( bcp_dat );
	    }

	    /*-- Create new data --*/
	    if ((bcp_dat = bcp_data_bind( bcp_cmd, result_type )) == NULL)
		goto return_fail;

	    while ((return_code = 
		    bcp_data_xfer( bcp_dat, 
				   bcp_cmd,
				   bcp_desc )) != CS_END_DATA)
	    {
		if (sg_interrupted)
		    goto return_interrupt;

		if (return_code != CS_SUCCEED)
		{
		    g_printerr( "bcp: bcp_data_xfer failed...\n");

		    if (++nerrors == maxerrors)
			goto return_fail;
		}
		else
		{
		    ++rows_in_batch;
		}

		if (rows_in_batch == batchsize)
		{
		    if (blk_done( bcp_desc,
				  CS_BLK_BATCH,
				  &nrows ) != CS_SUCCEED)
		    {
			g_printerr("bcp: blk_done failed!\n");

			if (++nerrors == maxerrors)
			    goto return_fail;
		    }
		    else
		    {
			total_rows += nrows;
			
			g_printerr("%d rows successfully bulk-copied to SQL Server.\n",
				   (int)nrows);
		    }

		    if (sg_interrupted)
			goto return_interrupt;

		    rows_in_batch = 0;
		}
	    } /* while (bcp_data_xfer()) */

	  default:
	    break;
	} /* switch (result_type) */

    }
#if 0
    if (rows_in_batch > 0) 
    {
	if (blk_done( bcp_desc,
		      CS_BLK_BATCH,
		      &nrows ) != CS_SUCCEED)
	{
	    goto return_fail;
	}
	
	if (sg_interrupted)
	    goto return_interrupt;
	
	if (nrows != rows_in_batch) 
	{
	    if (++nerrors == maxerrors) 
		goto return_fail;
	}
	total_rows += nrows;
    }
#endif

#if defined DEBUG
    g_printerr("bcp: blk_done(CS_BLK_ALL)\n");
#endif
    if (blk_done( bcp_desc,
		  CS_BLK_ALL,
		  &nrows ) != CS_SUCCEED)
	goto return_fail;

    total_rows += nrows;

    gettimeofday( &tv_end, NULL );
	
    if (nrows != 0)
	g_printerr("%d rows successfully bulk-copied to SQL Server\n",
		   (int)nrows);

    g_printerr( "\n%d row%s copied.\n", total_rows, 
		(total_rows != 1) ? "s" : "" );

    secs = ELAPSED_SEC(tv_start,tv_end);
    g_printerr("Clock Time (sec.): Total = %-.4f  Avg = %-.4f (%.2f rows per sec.)\n",
	       secs, secs / (double)total_rows, (double)total_rows / secs );
	
    return_code = 0; //CMD_CLEARBUF;
    goto leave;

 return_interrupt:
    if (bcp_desc != NULL)
	blk_done( bcp_desc, CS_BLK_CANCEL, &nrows );
    ct_cancel( bcp->connection, (CS_COMMAND*)NULL, CS_CANCEL_ALL );
    ct_cancel( dbh->connection, (CS_COMMAND*)NULL, CS_CANCEL_ALL );

    return_code = 0; //CMD_CLEARBUF;
    goto leave;

 return_fail:
    if (bcp->connection != NULL)
    {
	if (ct_con_props( bcp->connection,        /* Connection */
			  CS_GET,                 /* Action */
			  CS_CON_STATUS,          /* Property */
			  (CS_VOID*)&con_status,  /* Buffer */
			  CS_UNUSED,              /* Buffer Length */
			  (CS_INT*)NULL ) != CS_SUCCEED)
	{
	    con_status = CS_CONSTAT_CONNECTED;
	}

	/*-- If connected, disconnect --*/
	if (con_status == CS_CONSTAT_CONNECTED)
	{  
	    if (bcp_desc != NULL)
	    {
		blk_done( bcp_desc, CS_BLK_CANCEL, &nrows );
	    }
	    
	    ct_cancel( bcp->connection, (CS_COMMAND*)NULL, CS_CANCEL_ALL );
	}
    }

    if (dbh->connection != NULL)
    {
	if (ct_con_props( dbh->connection,           /* Connection */
			  CS_GET,                 /* Action */
			  CS_CON_STATUS,          /* Property */
			  (CS_VOID*)&con_status,  /* Buffer */
			  CS_UNUSED,              /* Buffer Length */
			  (CS_INT*)NULL ) != CS_SUCCEED)
	{
	    con_status = CS_CONSTAT_CONNECTED;
	}

	if (con_status == CS_CONSTAT_CONNECTED)
	{
	    ct_cancel( dbh->connection, (CS_COMMAND*)NULL, CS_CANCEL_ALL );
	}
    }

    return_code = -1; //CMD_FAIL;

 leave:
    if (bcp_dat != NULL)
	bcp_data_destroy( bcp_dat );

    if (bcp_cmd != NULL)
	ct_cmd_drop( bcp_cmd );
	
    if (bcp_desc != NULL)
    {
	blk_drop( bcp_desc );
    }
    value = 0;
    ct_options(dbh->connection, 
	       CS_SET,
	       CS_OPT_ROWCOUNT,
	       &value,
	       CS_UNUSED,
	       NULL);


    return return_code;
}

static bcp_data_t* 
bcp_data_bind ( CS_COMMAND *cmd, CS_INT result_type )
{
    CS_INT        ncols;
    CS_INT        i;
    bcp_data_t   *d;
    bcp_col_t    *c;

    /*-- Retrieve the number of columns in the result set --*/
    if (ct_res_info( cmd,               /* Command */
		     CS_NUMDATA,        /* Type */
		     (CS_VOID*)&ncols,  /* Buffer */
		     CS_UNUSED,         /* Buffer Length */
		     (CS_INT*)NULL) != CS_SUCCEED)
    {
	g_printerr( "bcp_data_bind: Unable to fetch number of columns\n" );
	return NULL;
    }

    d = g_new(bcp_data_t, 1);
    c = g_new(bcp_col_t,  ncols );

    if (d == NULL || c == NULL) 
    {
	g_printerr( "bcp_data_bind: Memory allocation failure.\n" );
	if (d != NULL)
	    g_free( d );
	if (c != NULL)
	    g_free( c );
	return NULL;
    }

    /*-- Inialize d --*/
    d->d_type  = result_type;
    d->d_ncols = ncols;
    d->d_cols  = c;

    for (i = 0; i < ncols; i++) 
	d->d_cols[i].c_data = NULL;
	
    for (i = 0; i < ncols; i++)
    {
	c = &d->d_cols[i];
	c->c_colid   = i + 1;
	c->c_len     = -1;
	c->c_nullind = 0;

	/*-- Get description for column --*/
	if (ct_describe( cmd, i+1, &c->c_format ) != CS_SUCCEED) 
	{
	    g_printerr("bcp_data_bind: Unable to fetch column %d description\n",
		       (int) i+1 );
	    bcp_data_destroy( d );
	    return NULL;
	}

	/*-- Allocate enough space to hold data --*/
	c->c_data = g_new(char, c->c_format.maxlength );

	/*-- Clean up format --*/
	c->c_format.count  = 1;
	c->c_format.locale = NULL;

	/*-- Bind to the data space --*/
	if (ct_bind( cmd,                            /* Command */
		     i + 1,                          /* Item */
		     &c->c_format,                   /* Format */
		     (CS_VOID*)c->c_data,            /* Buffer */
		     (CS_INT*)&c->c_len,             /* Data Copied */
		     &c->c_nullind                   /* NULL Indicator */
		) != CS_SUCCEED)
	{
	    g_printerr("bcp_data_bind: Unable to bind column %d\n",
		       (int) i+1 );
	    bcp_data_destroy( d );
	    return NULL;
	}
    }

    return d;
}

static CS_INT 
bcp_data_xfer( bcp_data_t *d, CS_COMMAND *cmd, CS_BLKDESC *blkdesc )
{
    CS_RETCODE  return_code;
    CS_INT      nrows;
    CS_INT      i;
/*    CS_INT      col; */
    bcp_col_t  *c;

    return_code = ct_fetch( cmd,              /* Command */
			    CS_UNUSED,        /* Type */
			    CS_UNUSED,        /* Offset */
			    CS_UNUSED,        /* Option */
			    &nrows );
	
    if (return_code != CS_SUCCEED)
    {
	return return_code;
    }

    for (i = 0; i < d->d_ncols; i++)
    {
	c = &d->d_cols[i];

	/*
	 * Here we determine whether or not we need to (re-) bind to
	 * the incoming data.  If the length of the data has changed
	 * since the last row, or if it has become NULL, then we
	 * want to re-bind.
	 */
	if (c->c_len == -1 || c->c_len != c->c_prevlen ||
	    c->c_nullind != c->c_prevnullind)
	{
	    
	    if (c->c_nullind == -1)
	    {
		c->c_len = 0;
	    }

	    /* no id-insert, and target column is identity */
	    if (!sg_table->id_insert && 
		sg_table->target_has_identity && 
		i == sg_table->target_id_col) 
	    {
#if defined DEBUG
		g_printerr("bcp: column %d is IDENTITY, skipping\n", i+1);
#endif
		c->c_nullind = 0;
		c->c_len = 0;
		continue;
	    }
#if defined DEGUG
	    g_printerr("bcp: binding column %d, c_len=%d, c_nullind=%d\n",
		       i+1, c->c_len, c->c_nullind);
#endif

	    c->c_prevlen     = c->c_len;
	    c->c_prevnullind = c->c_nullind;

	    if (blk_bind( blkdesc,              /* Block Descriptor */
			  i + 1,                /* Column Number */
			  &c->c_format,         /* Data Format */
			  (CS_VOID*)c->c_data,  /* Buffer */
			  &c->c_len,        		/* Buffer Length */
			  &c->c_nullind ) != CS_SUCCEED)
	    {
		g_printerr("bcp_data_xfer: Unable to bind results for column %d\n", 
			   (int) i+1 );
		return CS_FAIL;
	    }
	}
    }

#if defined DEBUG
    dump_row(d);
#endif

    if (blk_rowxfer( blkdesc ) != CS_SUCCEED)
    {
	g_printerr("bcp_data_xfer: Unable to transfer row to remote SQL Server\n" );
	return CS_FAIL;
    }

    return CS_SUCCEED;
}

/*
 * bcp_data_destroy():
 *
 * Destroys a bcp_data_t structure.
 */
static void 
bcp_data_destroy( bcp_data_t *d )
{
    CS_INT  i;

    if (d != NULL)
    {
	if (d->d_cols != NULL)
	{
	    for (i = 0; i < d->d_ncols; i++)
	    {
		if (d->d_cols[i].c_data != NULL)
		    g_free( d->d_cols[i].c_data );
	    }

	    g_free( d->d_cols );
	}

	g_free( d );
    }
}

/*
 * bcp_signal():
 *
 * This function is called whenever a signal is received, its
 * only real job is to register the fact that a signal was caught
 * during processing via the sg_interrupted flag.
 */
static void 
bcp_signal( int sig, void *user_data )
{
    sg_interrupted = TRUE;

    if (sg_bcp_connection != NULL)
    {
	ct_cancel( sg_bcp_connection, (CS_COMMAND*)NULL, CS_CANCEL_ATTN );
    }
    ct_cancel( sg_dbh_connection, (CS_COMMAND*)NULL, CS_CANCEL_ALL );
}

/*
 * build_sql()
 * Create the select statement based on the selected columns
 */
static char *
build_sql(TABLE *table)
{
    static char sql_buf[1024];
    int i;
    int first = 1;

    strcpy(sql_buf, "select ");
    for (i = 0; i < table->numCols; ++i) {
	if (table->cols[i]->selected) {
	    if (!first)
		strcat(sql_buf, ", ");
	    strcat(sql_buf, table->cols[i]->name);
	    first = 0;
	}
    }
    strcat(sql_buf, "\n from ");
    strcat(sql_buf, table->name);
    if(table->where_clause) {
	strcat(sql_buf, "\n");
	strcat(sql_buf, table->where_clause);
    }

    g_printerr("bcp: SQL command for query: %s\n", sql_buf);
    return sql_buf;
}

/*
 * get_target_info()
 * Get target table schema info - currently used to determine if the
 * target table has an identity column
 */
static void
get_target_info (DBH *bcp, TABLE *table)
{
    char   sql[512];
    RESULT *ret;
    ROW    *row;
    int    i;
    char *name = table->target_table[0] ? table->target_table : table->name;
    char *dbname = table->target_db[0] ? table->target_db : table->dbname;

    sprintf(sql, "
select distinct Column_name = c.name, 
   Type = t.name, 
   Length = c.length, 
   Prec = c.prec, 
   Scale = c.scale, 
   Nulls = convert(bit, (c.status & 8)),
   Ident = convert(bit, (c.status & 0x80))
from   %s.dbo.syscolumns c, %s.dbo.systypes t
where  c.id = object_id('%s.dbo.%s')
and    c.usertype *= t.usertype
order by c.colid
",
	    dbname, dbname, dbname, name);
    
    ret = db_exec_sql(bcp, sql);

    g_return_if_fail(ret != NULL);
    g_return_if_fail(ret->result != NULL);

    for (row = ret->result->row, i = 0; row; row = row->next, ++i) {
	if (row->data[6].value.i) {
	    table->target_has_identity = 1;
	    table->target_id_col = i;
	    table->target_id_precision = row->data[3].value.i;
	}
    }

    db_free_result(ret);
}

#if 0
/*
 * bcp_server_cb():
 *
 * Sybase callback message handler for all messages coming from the
 * SQL Server or Open Server.
 *
 */
static CS_RETCODE bcp_server_cb (CS_CONTEXT    *ctx, 
				 CS_CONNECTION *con, 
				 CS_SERVERMSG *msg)
{
    /*
     * Ignore "database changed", or "language changed" messages from
     * the server.
     */
    if( msg->msgnumber == 5701 ||    /* database context change */
	msg->msgnumber == 5703 ||    /* language changed */
	msg->msgnumber == 5704 )     /* charset changed */
    {
	return CS_SUCCEED;
    }

    if (msg->severity >= 0)
    {
	/*
	 * If the severity is something other than 0 or the msg number is
	 * 0 (user informational messages).
	 */
	if (msg->severity >= 0 || msg->msgnumber == 0) 
	{
	    /*
	     * If the message was something other than informational, and
	     * the severity was greater than 0, then print information to
	     * stderr with a little pre-amble information.  According to 
	     * the Sybase System Administrator's guide, severity level 10
	     * messages should not display severity information.
	     */
	    if (msg->msgnumber > 0) // && msg->severity > 10)
	    {
		g_printerr(  "Msg %d, Level %d, State %d\n",
			     (int)msg->msgnumber, (int)msg->severity, (int)msg->state );
		if (msg->svrnlen > 0)
		    g_printerr(  "Server '%s'", (char*)msg->svrname );
		if (msg->proclen > 0)
		    g_printerr(  ", Procedure '%s'", (char*)msg->proc );
		if( msg->line > 0 )
		    g_printerr(  ", Line %d", (int)msg->line );
		g_printerr(  "\n" );
		g_printerr(  "%s\n", msg->text );
//		fflush( stderr );
	    }
	    else 
	    {
		/*
		 * Otherwise, it is just an informational (e.g. print) message
		 * from the server, so send it to stdout.
		 */
		g_print( "%s\n", msg->text );
	    }
	}
    }
    
    return CS_SUCCEED;
}

/*
 * bcp_client_cb():
 */
static CS_RETCODE bcp_client_cb ( CS_CONTEXT    *ctx, 
				  CS_CONNECTION *con, 
				  CS_CLIENTMSG  *msg )
{
    /*
     * Let the CS-Lib handler print the message out.
     */
    g_printerr( "Open Client Message\n" );
    if (CS_NUMBER(msg->msgnumber) > 0)
    {
	g_printerr("Layer %d, Origin %d, Severity %d, Number %d\n",
		  (int)CS_LAYER(msg->msgnumber),
		  (int)CS_ORIGIN(msg->msgnumber),
		  (int)CS_SEVERITY(msg->msgnumber),
		  (int)CS_NUMBER(msg->msgnumber) );
    }
    if (msg->osstringlen > 0)
    {
	g_printerr( "OS Error: %s\n", (char*)msg->osstring );
    }
    g_printerr(  "%s\n", msg->msgstring );
//	fflush( stderr ) ;

    return CS_SUCCEED ;
}
#endif

#if defined DEBUG
static void
dump_row( bcp_data_t *d)
{
    bcp_col_t *c;
    char buff[255];
    int i;
    CS_DATAFMT destfmt;
    CS_INT reslen;

    memset(&destfmt, 0, sizeof(destfmt));
    destfmt.datatype = CS_CHAR_TYPE;
    destfmt.maxlength = 255;
    destfmt.format   = CS_FMT_NULLTERM;

    for (i = 0; i < d->d_ncols; i++) {
	c = &d->d_cols[i];

	if(cs_convert(sg_context, &c->c_format, c->c_data,
		   &destfmt, buff, &reslen) != CS_SUCCEED) {
	    g_printerr("bcp: cs_convert failed\n");
	    return;
	}
	printf("%.*s\t", reslen, buff);
    }
    printf("\n");
}

#endif	
    
    
