/*
 * $Id: callbacks.c,v 1.2 2001/01/17 19:43:33 mpeppler Exp $
 *
 * callbacks.c   - UI handling
 *
 * Copyright (C) 2001 by Michael Peppler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * You may contact the author :
 *   e-mail:  mpeppler@peppler.org
 *
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include "bcptool.h"

#include "bcp.h"


static void
fill_dblist (void);
static void
set_clist_number(GtkWidget *clist, int row, int col, int value);
static int
set_dbopt (DBH *dbh, char *db, char *option, char *val);


static DB **databases;
static int num_dbs;
static int cur_db;
static int cur_table;

/* This value gets set to 1 when the table option window is opened
   so that the programatical selection of clist items does not fire
   the callback and mess up the count of selected columns in the
   user table window... */
static int no_count;

static GtkWidget *errLog;

static DBH *dbh, *bcp;

static FILE *log_fh;	/* if set, log all errors/messages to this handle */

static int
set_dbopt (DBH *dbh, char *db, char *option, char *val)
{
    RESULT *ret;
    char sql[256];

    db_use(dbh, "master");
    sprintf(sql, "exec sp_dboption %s, '%s', %s",
	    db, option, val);
    ret = db_exec_sql(dbh, sql);
    if(!ret || ret->procStatus) {
	g_printerr("Can't set option %s on %s to %s (proc returned %d)\n",
		   option, db, val, ret->procStatus);
	goto error;
    }
    db_use(dbh, db);
    ret = db_exec_sql(dbh, "checkpoint");
    if(!ret || ret->procStatus) {
	g_printerr("Can't set option %s on %s to %s (proc returned %d)\n",
		   option, db, val, ret->procStatus);
	goto error;
    }

    db_use(dbh, "master");
    db_free_result(ret);

    return 0;
    

 error:
    if(ret)
	db_free_result(ret);
    return -1;
}

static void
load_table_info (TABLE *table, char *dbname)
{
    char   sql[512];
    RESULT *ret;
    ROW    *row;
    int    i;

    sprintf(sql, "
select distinct Column_name = c.name, 
   Type = t.name, 
   Length = c.length, 
   Prec = c.prec, 
   Scale = c.scale, 
   Nulls = convert(bit, (c.status & 8)),
   Ident = convert(bit, (c.status & 0x80))
from   %s.dbo.syscolumns c, %s.dbo.systypes t
where  c.id = object_id('%s.dbo.%s')
and    c.usertype *= t.usertype
order by c.colid
",
	    dbname, dbname, dbname, table->name);

    table->dbname = dbname;
    
    ret = db_exec_sql(dbh, sql);

    g_return_if_fail(ret != NULL);
    g_return_if_fail(ret->result != NULL);

    if(ret->result->last) {
	table->numCols = ret->result->last->rowNum;
    } else if(ret->result->row) {
	table->numCols = ret->result->row->rowNum;
    } else {
	table->numCols = 0;
    }
    table->numSelectedCols = table->numCols;

    if(table->numCols) {
	table->cols = g_new0(COLUMN *, table->numCols);
	row = ret->result->row;
    } else {
	row = NULL;
    }
    
    i = 0;
    while(row) {
	table->cols[i] = g_new0(COLUMN, 1);
	strcpy(table->cols[i]->name, 
	       row->data[0].value.c);
	strcpy(table->cols[i]->type_name, 
	       row->data[1].value.c);
	table->cols[i]->length    = row->data[2].value.i;
	table->cols[i]->precision = row->data[3].value.i;
	table->cols[i]->scale     = row->data[4].value.i;
	table->cols[i]->nullable  = row->data[5].value.i;
	table->cols[i]->identity  = row->data[6].value.i;
	table->cols[i]->selected  = 1;

//	load_table_info(db->tables[i]);
	
	row = row->next;
	++i;
    }
    db_free_result(ret);

    db_use(dbh, dbname);

    sprintf(sql, "select sum(rowcnt(doampg))
                , max(indid)
                , max(indid * abs(sign(indid - 255)))
                , sum(reserved_pgs(id, doampg))
                , sum(reserved_pgs(id, ioampg))
           from sysindexes where id = object_id('%s')",
	    table->name);

    ret = db_exec_sql(dbh, sql);
    if(ret->result)
	row = ret->result->row;
    else
	row = NULL;
    if(row) {
//	for(i = 0; i < 5; ++i)
//	    printf("row[%d] = %d\n", i, row->data[i].value.i);
	table->rows     = atoi(row->data[0].value.c);
	table->tiindex  = row->data[1].value.i;
	table->maxindex = row->data[2].value.i;
	table->pages    = row->data[3].value.i;
	table->tipages  = row->data[4].value.i;
    }
    db_free_result(ret);

    db_use(dbh, "master");
}

static void 
load_db_info (DB *db)
{
    char   sql[512];
    RESULT *ret;
    ROW    *row;
    int    i;

    sprintf(sql, "select name from %s..sysobjects where type = 'U'",
	    db->name);
    
    ret = db_exec_sql(dbh, sql);

    if(!ret->result)
	return;

    row = ret->result->row;

    if(ret->result->last) {
	db->num_tables = ret->result->last->rowNum;
    } else if(ret->result->row) {
	db->num_tables = ret->result->row->rowNum;
    } else {
	db->num_tables = 0;
    }

    if(db->num_tables) {
	db->tables = g_new0(TABLE *, db->num_tables);
    }
    
    i = 0;
    while(row) {
	db->tables[i] = g_new0(TABLE, 1);
	strcpy(db->tables[i]->name, 
	       row->data[0].value.c);
	load_table_info(db->tables[i], db->name);
	
	row = row->next;
	++i;
    }
    db_free_result(ret);

    set_clist_number(lookup_widget(window1, "clist1"), 
		     cur_db, 1, db->num_tables);
    
}

static void
fmt_options_string(DB* db, char *options)
{
    if(db->options == 0)
	strcpy(options, "None"); // XXX
    else {
	options[0] = 0;
	if(db->options & DBOPT_BCP)
	    strcat(options, "BCP");
	if(db->options & DBOPT_TRUNC) {
	    if(options[0])
		strcat(options, ",");
	    strcat(options, "TLC");
	}
    }
}

static void
set_option(int opt)
{
    char options[20];
    GtkWidget *widget = lookup_widget(window1, "clist1");

    databases[cur_db]->options ^= opt;
    fmt_options_string(databases[cur_db], options);

//    printf("options: %s\n", options);

    gtk_clist_set_text(GTK_CLIST(widget),
		       cur_db, 3, options);
}

static void
set_clist_number(GtkWidget *clist, int row, int col, int value)
{
    char buff[20];

//    databases[cur_db]->options ^= opt;
//    fmt_options_string(databases[cur_db], options);

    sprintf(buff, "%d", value);

    gtk_clist_set_text(GTK_CLIST(clist),
		       row, col, buff);
}

static void
set_selected_fields(void)
{
    GtkWidget *widget = lookup_widget(window3, "clist2");
    char fmt[20];

    sprintf(fmt, "%3d / %3d",
	    databases[cur_db]->tables[cur_table]->numCols,
	    databases[cur_db]->tables[cur_table]->numSelectedCols);

    gtk_clist_set_text(GTK_CLIST(widget),
		       cur_table, 1, fmt);
}
    



static void
printError(char *string)
{
    if(!errLog) {
	errLog = lookup_widget(window1, "text1");
    }

    gtk_text_insert( GTK_TEXT(errLog), NULL, NULL, NULL,
		     string, -1);
    if(log_fh) {
	fprintf(log_fh, string);
	fflush(log_fh);
    }
}


static void
fill_dblist (void)
{
    char *strings[4];
    char tables[10], selected[10], options[30];
    GtkWidget *clist = lookup_widget(window1, "clist1");
    int i;

    gtk_clist_clear(GTK_CLIST(clist));

    for(i = 0; i < num_dbs; ++i) {
	strings[0] = databases[i]->name;
	sprintf(tables, "%3d", databases[i]->num_tables);
	sprintf(selected, "%3d", databases[i]->selected);

	fmt_options_string(databases[i], options);
		
	strings[1] = tables;
	strings[2] = selected;
	strings[3] = options;

	gtk_clist_append(GTK_CLIST(clist), strings);
    }

}

static void
fill_tablelist (void)
{
    char *strings[4];
    char fields[20], selected[10], rows[30];
    GtkWidget *clist = lookup_widget(window3, "clist2");
    int i;

    gtk_clist_clear(GTK_CLIST(clist));

    for(i = 0; i < databases[cur_db]->num_tables; ++i) {
	strings[0] = databases[cur_db]->tables[i]->name;
	sprintf(fields, "%3d / %3d", databases[cur_db]->tables[i]->numCols,
		databases[cur_db]->tables[i]->numSelectedCols);
	sprintf(selected, "%s", 
		databases[cur_db]->tables[i]->selected ? "Yes" : "No");
	sprintf(rows, "%10d", databases[cur_db]->tables[i]->rows);
		
	strings[1] = fields;
	strings[2] = rows;
	strings[3] = selected;

	gtk_clist_append(GTK_CLIST(clist), strings);
    }

}

static void
fill_collist (void)
{
    char *strings[4];
    char length[20], nullable[10], selected[30];
    GtkWidget *clist = lookup_widget(window4, "clist3");
    int i;
    TABLE *tbl = databases[cur_db]->tables[cur_table];
    GtkWidget *widget;
    char buff[20];

    gtk_clist_clear(GTK_CLIST(clist));

    for(i = 0; i < tbl->numCols; ++i) {
	strings[0] = tbl->cols[i]->name;
	sprintf(length, "%d", tbl->cols[i]->length);
	sprintf(nullable, "%s", tbl->cols[i]->nullable ? "Yes" : "No");
	sprintf(selected, "%s", 
		tbl->cols[i]->selected ? "Yes" : "No");
		
	strings[1] = length;
	strings[2] = nullable;
	strings[3] = selected;

	gtk_clist_append(GTK_CLIST(clist), strings);

	no_count = 1;
	
	if(tbl->cols[i]->selected) {
	    gtk_clist_select_row(GTK_CLIST(clist), i, 0);
	}

	no_count = 0;
    }

// NOT USED YET
//    widget = lookup_widget(window4, "index_entry");
//    gtk_entry_set_text(GTK_ENTRY(widget), "0");
    widget = lookup_widget(window4, "rowcount_entry");
    sprintf(buff, "%d", tbl->rowcount);
    gtk_entry_set_text(GTK_ENTRY(widget), buff);
    widget = lookup_widget(window4, "batch_size_entry");
    if(!tbl->batchsize) {
	gtk_entry_set_text(GTK_ENTRY(widget), "1000");
    } else {
	sprintf(buff, "%d", tbl->batchsize);
	gtk_entry_set_text(GTK_ENTRY(widget), buff);
    }

    widget = lookup_widget(window4, "target_table_entry");
    gtk_entry_set_text(GTK_ENTRY(widget), tbl->target_table);

    widget = lookup_widget(window4, "target_db_entry");
    gtk_entry_set_text(GTK_ENTRY(widget), tbl->target_db);

    widget = lookup_widget(window4, "where_clause_entry");
    gtk_entry_set_text(GTK_ENTRY(widget), tbl->where_clause);

    widget = lookup_widget (window4, "trunc_target_check");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget),
				  tbl->trunc_target);

    widget = lookup_widget (window4, "identity_insert_check");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget),
				  tbl->id_insert);

    widget = lookup_widget (window4, "table_selected_check");
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget),
				  tbl->selected);
}





void
on_about_btn_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
    gtk_widget_show (dialog1);
}


void
on_setup_btn_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
    gtk_widget_show (window2);
}


void
on_start_btn_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
    int i, j;
    char bcp_table[100];
    int err;

    if(!dbh || !bcp) {
	g_printerr("Not connected to any databases!\n");
	return;
    }
    
    
    for(i = 0; i < num_dbs; ++i) {
	db_use(dbh, databases[i]->name);

	for(j = 0; j < databases[i]->num_tables; ++j) {
	    if(databases[i]->tables[j]->selected) {
		char *db = databases[i]->name;
		if(databases[i]->tables[j]->target_db[0])
		    db = databases[i]->tables[j]->target_db;

		/* set the database options for the target db
		   if necessary */

		if(databases[i]->options & DBOPT_BCP)
		    if(set_dbopt(bcp, db, "bulk", "true"))
			return;
		if(databases[i]->options & DBOPT_TRUNC)
		    if(set_dbopt(bcp, db, "trunc", "true"))
			return;


		/* create the target table string - defaults to
		   the same db and table names as the source,
		   but this can be overridden in the UI */
		sprintf(bcp_table, "%s..%s", db,
			databases[i]->tables[j]->target_table[0] ? 
			databases[i]->tables[j]->target_table :
			databases[i]->tables[j]->name);
		g_printerr("Processing table %s\n", bcp_table);
		if((err = cmd_bcp(dbh, bcp, databases[i]->tables[j], bcp_table))) {
		    g_printerr("BCP command failed - aborted\n");
		}
		/* reset the db options */

		if(databases[i]->options & DBOPT_BCP)
		    if(set_dbopt(bcp, db, "bulk", "false"))
			++err;
		if(databases[i]->options & DBOPT_TRUNC)
		    if(set_dbopt(bcp, db, "trunc", "false"))
			++err;

		if(err) {
		    return;
		}
	    }
	}
    }
}


void
on_quit_btn_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
    gtk_main_quit();
}


void
on_setup_apply_btn_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
    GtkWidget *widget;
    char *server;
    char *user;
    char *pwd;
    char *charset;
    char *packetsize;
    int  encrypt;
    char *where;
    char sql[512];
    RESULT *ret;
    ROW *row;
    int i;
    char *logfile;

    db_set_errfunc(printError);
    db_set_msgfunc(printError);

    g_set_print_handler(printError);
    g_set_printerr_handler(printError);

    widget = lookup_widget(window2, "source_server");
    server = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "source_login");
    user = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "source_password");
    pwd = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "source_charset");
    charset = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "source_packet_size");
    packetsize = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "source_encrypt_btn");
    encrypt = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON ( widget ) );

    widget = lookup_widget(window2, "log_file");
    logfile = gtk_entry_get_text( GTK_ENTRY ( widget ) );

    if(logfile && *logfile) {
	log_fh = fopen(logfile, "a");
	if(!log_fh)
	    g_printerr("Can't open %s: %s\n", logfile, strerror(errno));
    }


    if(!*server || !*user) {
	widget = lookup_widget(dialog2, "error_label");
	gtk_label_set_text ( GTK_LABEL (widget), "Please enter a source server name, a target server name, a source login name and a target login name.\nPasswords and other information can be entered on an as needed basis");
	gtk_widget_show( dialog2 );
	return;
    }

    
    
    if(dbh)
	db_free(dbh);

    dbh = db_connect(server, user, pwd, packetsize, encrypt, charset, 0);
    if(!dbh) {
	/* XXX - popup an error window? */
	return;
    }
#if defined DEBUG
    dbh->debug = 3;
#endif

    widget = lookup_widget(window2, "target_server");
    server = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "target_login");
    user = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "target_password");
    pwd = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "target_charset");
    charset = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "target_packet_size");
    packetsize = gtk_entry_get_text(GTK_ENTRY(widget));
    widget = lookup_widget(window2, "target_encrypt_btn");
    encrypt = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON ( widget ) );

    if(!*server || !*user) {
	widget = lookup_widget(dialog2, "error_label");
	gtk_label_set_text ( GTK_LABEL (widget), "Please enter a source server name, a target server name, a source login name and a target login name.\nPasswords and other information can be entered on an as needed basis");
	gtk_widget_show( dialog2 );
	return;
    }

    
    if(bcp)
	db_free(bcp);

    bcp = db_connect(server, user, pwd, packetsize, encrypt, charset, 1);
    if(!bcp) {
	/* XXX - popup an error window? */
	return;
    }


    if(num_dbs) {
	for(i = 0; i < num_dbs; ++i)
	    g_free(databases[i]);
	g_free(databases);
    }

    widget = lookup_widget(window2, "where_clause");
    where = gtk_entry_get_text(GTK_ENTRY(widget));
    
    if(where && *where)
	sprintf(sql, "select name from master..sysdatabases where %s\n",
		where);
    else 
	strcpy(sql, "select name from master..sysdatabases");

    
    ret = db_exec_sql(dbh, sql);

    if(!ret->result)
	return;

    row = ret->result->row;

    if(ret->result->last) {
	num_dbs = ret->result->last->rowNum;
    } else if(ret->result->row) {
	num_dbs = ret->result->row->rowNum;
    } else {
	num_dbs = 0;
    }

    if(num_dbs) {
	databases = g_new0(DB *, num_dbs);
    }
    
    i = 0;
    while(row) {
	databases[i] = g_new0(DB, 1);
	strcpy(databases[i]->name, row->data[0].value.c);

	row = row->next;
	++i;
    }
    db_free_result(ret);
    fill_dblist();

    gtk_widget_hide(window2);

    widget = lookup_widget(window1, "start_btn");
    gtk_widget_set_sensitive(widget, 1);
}


void
on_setup_reset_btn_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_about_ok_btn_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
    gtk_widget_hide(dialog1);
}


gboolean
on_window1_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    gtk_main_quit();
    return FALSE;
}


gboolean
on_window2_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    gtk_widget_hide(widget);
    return TRUE;
}


gboolean
on_dialog1_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    gtk_widget_hide(widget);
    return TRUE;
}


void
on_clist1_select_row                   (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    cur_db = row;

    if(!databases[cur_db]->num_tables)
	load_db_info(databases[cur_db]);

    fill_tablelist();

    gtk_widget_show(window3);
}


void
on_clist2_select_row                   (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    cur_table = row;

    fill_collist();

    gtk_widget_show(window4);
}


gboolean
on_window3_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    gtk_widget_hide(widget);
    return TRUE;
}


void
on_bcp_check_toggled                   (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
    set_option(DBOPT_BCP);
}


void
on_trunc_log_check_toggled             (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
    set_option(DBOPT_TRUNC);
}



gboolean
on_window4_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    gtk_widget_hide(widget);

    return TRUE;
}


void
on_table_selected_check_toggled        (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{

}


void
on_table_apply_btn_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
    GtkWidget *widget;
    char *p;
    TABLE *table = databases[cur_db]->tables[cur_table];
    int selected;
    int i;

    widget = lookup_widget(window4, "rowcount_entry");
    p = gtk_entry_get_text( GTK_ENTRY ( widget ) );
    if(p && *p)
	table->rowcount = atoi(p);

    widget = lookup_widget(window4, "batch_size_entry");
    p = gtk_entry_get_text( GTK_ENTRY ( widget ) );
    if(p && *p)
	table->batchsize = atoi(p);
    
    widget = lookup_widget(window4, "target_table_entry");
    p = gtk_entry_get_text( GTK_ENTRY ( widget ) );
    if(p && *p)
	strcpy(table->target_table, p);

    widget = lookup_widget(window4, "target_db_entry");
    p = gtk_entry_get_text( GTK_ENTRY ( widget ) );
    if(p && *p)
	strcpy(table->target_db, p);

    widget = lookup_widget(window4, "where_clause_entry");
    p = gtk_entry_get_text( GTK_ENTRY ( widget ) );
    if(p && *p)
	strcpy(table->where_clause, p);

    widget = lookup_widget(window4, "target_db_entry");
    p = gtk_entry_get_text( GTK_ENTRY ( widget ) );
    if(p && *p)
	strcpy(table->target_db, p);

    widget = lookup_widget (window4, "trunc_target_check");
    table->trunc_target = 
	gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));

    widget = lookup_widget (window4, "identity_insert_check");
    table->id_insert = 
	gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));

    widget = lookup_widget (window4, "table_selected_check");
    table->selected = 
	gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));

    /* now update the display in the table selection window */
    widget = lookup_widget(window3, "clist2");

    if(databases[cur_db]->tables[cur_table]->selected) {
	gtk_clist_set_text(GTK_CLIST(widget), cur_table, 3, "Yes");
//	++databases[cur_db]->selected;
    }
    else {
	gtk_clist_set_text(GTK_CLIST(widget), cur_table, 3, "No");
//	--databases[cur_db]->selected;
    }

    /* Now update the "selected" column in the main window */
    for(i = 0, selected = 0; i < databases[cur_db]->num_tables; ++i) {
	if(databases[cur_db]->tables[i]->selected) {
	    ++selected;
	}
    }

    set_clist_number(lookup_widget(window1, "clist1"), cur_db, 2,
		     selected);


    gtk_widget_hide (window4);
}


void
on_table_reset_btn_clicked             (GtkButton       *button,
                                        gpointer         user_data)
{
    gtk_widget_hide (window4);
}


void
on_make_selection_btn_clicked          (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_clear_selection_btn_clicked         (GtkButton       *button,
                                        gpointer         user_data)
{
    int i;
    GtkWidget *widget = lookup_widget(window3, "clist2");

    for(i = 0; i < databases[cur_db]->num_tables; ++i) {
	databases[cur_db]->tables[i]->selected = 0;
	gtk_clist_set_text(GTK_CLIST(widget), i, 3, "No");
    }

    set_clist_number(lookup_widget(window1, "clist1"), cur_db, 2, 0);

    gtk_widget_hide(window3);
}


void
on_clist3_select_row                   (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    /* if no_count is set it means that we are programmatically
       setting these rows to "selected" state, so the processing
       is not necessary. */
    if(no_count)
	return ;

    databases[cur_db]->tables[cur_table]->cols[row]->selected = 1;
    databases[cur_db]->tables[cur_table]->numSelectedCols++;
    gtk_clist_set_text(clist, row, 3, "Yes");

    set_selected_fields();
}


void
on_clist3_unselect_row                 (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    /* if no_count is set it means that we are programmatically
       setting these rows to "selected" state, so the processing
       is not necessary. */
    if(no_count)
	return ;

    databases[cur_db]->tables[cur_table]->cols[row]->selected = 0;
    gtk_clist_set_text(clist, row, 3, "No");
    databases[cur_db]->tables[cur_table]->numSelectedCols--;

    set_selected_fields();
}


void
on_source_encrypt_btn_toggled          (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{

}


void
on_target_encrypt_btn_toggled          (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{

}


void
on_err_ok_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    gtk_widget_hide( dialog2 );
}


gboolean
on_dialog2_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    gtk_widget_hide(dialog2);
    
    return TRUE;
}

