/*
 * $Id: dbutil.c,v 1.2 2001/01/17 19:43:43 mpeppler Exp $
 *
 * dbutil.c     - Simple DB access using CT-lib
 *
 * Copyright (C) 2001 by Michael Peppler
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, write to the Free Software
 * Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * You may contact the author :
 *   e-mail:  mpeppler@peppler.org
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>

#include <ctpublic.h>
#include "dbutil.h"

static CS_CONTEXT *context;

static void (*errFunc)();
static void (*msgFunc)();

static void 
croak(char *string) 
{
    fprintf(stderr, "dbutil: %s\n", string);
    exit(1);
}

static void 
warn(char *string) 
{
    fprintf(stderr, "dbutil: %s\n", string);
}


static CS_RETCODE CS_PUBLIC
clientmsg_cb(CS_CONTEXT *context, CS_CONNECTION *connection, 
	     CS_CLIENTMSG *errmsg)
{
    char buff[1024];
    char oserr[512];
    
    if(errmsg->osstringlen > 0) 
	sprintf(oserr, "Operating System Error: %s",
		errmsg->osstring);
    else
	oserr[0] = 0;

    sprintf(buff, "OpenClient message: %s %s\n",
	    errmsg->msgstring, oserr);

    if(msgFunc)
	(*msgFunc)(buff);
    else
	fprintf(stderr, buff);

    return CS_SUCCEED;
}

static CS_RETCODE CS_PUBLIC
servermsg_cb(CS_CONTEXT *context, CS_CONNECTION *connection, 
	     CS_SERVERMSG *srvmsg)
{
//    CS_COMMAND	*cmd;
//    CS_RETCODE	retcode;
    char buff[1024];

    if(srvmsg->msgnumber == 5701 || srvmsg->msgnumber == 5703 
       || srvmsg->msgnumber == 5704)
	return CS_SUCCEED;

/*
    if((ct_con_props(connection, CS_GET, CS_USERDATA,
		     &imp_dbh, CS_SIZEOF(imp_dbh), NULL)) != CS_SUCCEED)
	croak("Panic: servermsg_cb: Can't find handle from connection");
*/
    if(srvmsg->msgnumber) {
	sprintf(buff, "Server message: %ld/%ld %s\n",
		srvmsg->msgnumber, srvmsg->severity, srvmsg->text);
    } else {
	sprintf(buff, "%s\n", srvmsg->text);
    }
    
    if(errFunc) {
	(*errFunc)(buff);
    } else {
	fprintf(stderr, buff);
    }
    
    return CS_SUCCEED;
}

static CS_COMMAND *
syb_alloc_cmd(CS_CONNECTION *connection) 
{
    CS_RETCODE retcode;
    CS_COMMAND *cmd;

    if((retcode = ct_cmd_alloc(connection, &cmd)) != CS_SUCCEED) {
	return NULL;
    }
    return cmd;
}
	    


void 
db_init (void) 
{
    CS_RETCODE	retcode;
    CS_INT	netio_type = CS_SYNC_IO;

    if((retcode = cs_ctx_alloc(CTLIB_VERSION, &context)) != CS_SUCCEED)
	croak("initialize: cs_ctx_alloc() failed");

    if((retcode = ct_init(context, CTLIB_VERSION)) != CS_SUCCEED)
    {
	cs_ctx_drop(context);
	context = NULL;
	croak("initialize: ct_init() failed");
    }

    if((retcode = ct_callback(context, NULL, CS_SET, CS_CLIENTMSG_CB,
			  (CS_VOID *)clientmsg_cb)) != CS_SUCCEED)
	croak("initialize: ct_callback(clientmsg) failed");
    if((retcode = ct_callback(context, NULL, CS_SET, CS_SERVERMSG_CB,
			      (CS_VOID *)servermsg_cb)) != CS_SUCCEED)
	croak("initialize: ct_callback(servermsg) failed");

    if((retcode = ct_config(context, CS_SET, CS_NETIO, &netio_type, 
			    CS_UNUSED, NULL)) != CS_SUCCEED)
	croak("initialize: ct_config(netio) failed");
    
}

void db_set_errfunc(void (*func)())
{
    errFunc = func;
}

void db_set_msgfunc(void (*func)())
{
    msgFunc = func;
}


DBH *
db_connect ( char *server, 
	     char *user, 
	     char *pwd,
	     char *packetsize,
	     int  encryption,
	     char *charset,
	     int  bcp ) 
{
    DBH *dbh;
    CS_INT retcode;

    dbh = calloc(1, sizeof(DBH));
    if(server)
	strcpy(dbh->server, server);
    if(user)
	strcpy(dbh->userName, user);
    if(pwd)
	strcpy(dbh->password, pwd);

    if((retcode = ct_con_alloc(context, &dbh->connection)) != CS_SUCCEED) {
	warn("ct_con_alloc failed");
	return NULL;
    }
    dbh->context = context;

    if(bcp) {	
	CS_BOOL           bcp_on = CS_TRUE; /* Flag to turn on bulk login */
	if (ct_con_props( dbh->connection,            /* Connection */
	                  CS_SET,                     /* Action */
	                  CS_BULK_LOGIN,              /* Property */
	                  (CS_VOID*)&bcp_on,          /* Buffer */
	                  CS_UNUSED,                  /* Buffer Lenth */
	                  (CS_INT*)NULL               /* Output Length */
						 ) != CS_SUCCEED)
	{
	    warn( "ct_con_props(CS_BULK_LOGIN) failed.\n");
	    return 0;
	}
    }


    if(retcode == CS_SUCCEED && dbh->userName[0]) {
	if((retcode = ct_con_props(dbh->connection, CS_SET, CS_USERNAME, 
				   dbh->userName, CS_NULLTERM, NULL)) != CS_SUCCEED)
	{
	    warn("ct_con_props(CS_USERNAME) failed");
	    return 0;
	}
	
    }
    if(retcode == CS_SUCCEED && dbh->password[0]) {
	if((retcode = ct_con_props(dbh->connection, CS_SET, CS_PASSWORD, 
				   dbh->password, CS_NULLTERM, NULL)) != CS_SUCCEED)
	{
	    warn("ct_con_props(CS_PASSWORD) failed");
	    return 0;
	}
    }

    if (ct_con_props( dbh->connection,             /* Connection */
		      CS_SET,                     /* Action */
		      CS_APPNAME,                 /* Property */
		      bcp ? (CS_VOID*)"bcptool-bcp":(CS_VOID*)"bcptool" ,
		      CS_NULLTERM,                /* Buffer Lenth */
		      (CS_INT*)NULL               /* Output Length */
	    ) != CS_SUCCEED)
    {
	warn("ct_con_props(CS_APPNAME) failed");
	return 0;
    }    

    if (packetsize != NULL) {
	int i = atoi(packetsize);
	if (ct_con_props( dbh->connection,         /* Connection */
			  CS_SET,                  /* Action */
			  CS_PACKETSIZE,           /* Property */
			  (CS_VOID*)&i,            /* Buffer */
			  CS_UNUSED,               /* Buffer Length */
			  (CS_INT*)NULL            /* Output Length */
		) != CS_SUCCEED)
	{
	    warn("ct_con_props(CS_PACKETSIZE) failed");
	    return 0;
	}
    }

    if (encryption) {
	int i = CS_TRUE;
	if (ct_con_props( dbh->connection,         /* Connection */
			  CS_SET,                  /* Action */
			  CS_SEC_ENCRYPTION,       /* Property */
			  (CS_VOID*)&i,            /* Buffer */
			  CS_UNUSED,               /* Buffer Length */
			  (CS_INT*)NULL            /* Output Length */
		) != CS_SUCCEED)
	{
	    warn("ct_con_props(CS_SEC_ENCRYPTION) failed\n");
	    return 0;
	}
    }

    /*
     * The following section initializes all locale type information.
     * First, we need to create a locale structure and initialize it
     * with information.
     */
    if (cs_loc_alloc( context, &dbh->locale ) != CS_SUCCEED)
    {
	warn( "cs_loc_alloc() failed.\n");
	return 0;
    }
    
    /*-- Initialize --*/
    if (cs_locale( context,                      /* Context */
		   CS_SET,                       /* Action */
		   dbh->locale,                  /* Locale Structure */
		   CS_LC_ALL,                    /* Property */
		   (CS_CHAR*)NULL,               /* Buffer */
		   CS_UNUSED,                    /* Buffer Length */
		   (CS_INT*)NULL                 /* Output Length */
	    ) != CS_SUCCEED)
    {
	warn("cs_locale(CS_LC_ALL) failed.\n");
	return 0;
    }

#if 0
    // Don't know if we need to support different languages
    /*-- Language --*/
    if( info->language != NULL && *info->language != '\0' ) {
	if (cs_locale( dbh->context,                 /* Context */
		       CS_SET,                    /* Action */
		       bcp_locale,                /* Locale Structure */
		       CS_SYB_LANG,               /* Property */
		       (CS_CHAR*)info->language,        /* Buffer */
		       CS_NULLTERM,               /* Buffer Length */
		       (CS_INT*)NULL              /* Output Length */
		) != CS_SUCCEED)
	{
	    g_printerr("bcp: Unable to set language to '%s' for BCP connection\n",
		       info->language );
	    goto return_fail;
	}
    }
#endif
    
    /*-- Character Set --*/
    if (charset != NULL && *charset != 0) {
	if (cs_locale( context,                   /* Context */
		       CS_SET,                    /* Action */
		       dbh->locale,               /* Locale Structure */
		       CS_SYB_CHARSET,            /* Property */
		       (CS_CHAR*)charset,         /* Buffer */
		       CS_NULLTERM,               /* Buffer Length */
		       (CS_INT*)NULL              /* Output Length */
		) != CS_SUCCEED)
	{
	    warn("cs_locale(CS_SYB_CHARSET) failed.\n");
	    return 0;
	}
    }
    
    /*-- Locale Property --*/
    if (ct_con_props( dbh->connection,         /* Connection */
		      CS_SET,                  /* Action */
		      CS_LOC_PROP,             /* Property */
		      (CS_VOID*)dbh->locale,   /* Buffer */
		      CS_UNUSED,               /* Buffer Length */
		      (CS_INT*)NULL            /* Output Length */
	    ) != CS_SUCCEED)
    {
	warn("ct_con_props(CS_LOC_PROP) failed\n");
	return 0;
    }


    
    if (retcode == CS_SUCCEED)
    {
	int len = dbh->server[0] == 0 ? 0 : CS_NULLTERM;
	if((retcode = ct_connect(dbh->connection, dbh->server, len)) != CS_SUCCEED) {
//	    cs_loc_drop(context, locale);
	    return 0;
	}
    }

    dbh->cmd = syb_alloc_cmd(dbh->connection);
    
    return dbh;
}

static void
clean_up(DBH *dbh) 
{
    int i;

    free(dbh->datafmt);
    for(i = 0; i < dbh->numCols; ++i) {
	if(dbh->coldata[i].type == CS_CHAR_TYPE ||
	   dbh->coldata[i].type == CS_IMAGE_TYPE)
	    free(dbh->coldata[i].value.c);
    }
    free(dbh->coldata);
}

static CS_CHAR * 
get_agg_op(CS_INT op)
{
    CS_CHAR *name;

    switch ((int)op)
    {
      case CS_OP_SUM:
	name = "sum";
	break;
      case CS_OP_AVG:
	name = "avg";
	break;
      case CS_OP_COUNT:
	name = "count";
	break;
      case CS_OP_MIN:
	name = "min";
	break;
      case CS_OP_MAX:
	name = "max";
	break;
      default:
	name = "unknown";
	break;
    }
    return name;
}

static CS_INT
get_cwidth(CS_DATAFMT *column)
{
    CS_INT		len;

    switch ((int) column->datatype)
    {
      case CS_CHAR_TYPE:
      case CS_VARCHAR_TYPE:
      case CS_TEXT_TYPE:
      case CS_IMAGE_TYPE:
	len = column->maxlength;
	break;

      case CS_BINARY_TYPE:
      case CS_VARBINARY_TYPE:
	len = (2 * column->maxlength) + 2;
	break;
	
      case CS_BIT_TYPE:
      case CS_TINYINT_TYPE:
	len = 3;
	break;
	
      case CS_SMALLINT_TYPE:
	len = 6;
	break;
	
      case CS_INT_TYPE:
	len = 11;
	break;
	
      case CS_REAL_TYPE:
      case CS_FLOAT_TYPE:
	len = 20;
	break;
	
      case CS_MONEY_TYPE:
      case CS_MONEY4_TYPE:
	len = 24;
	break;
	
      case CS_DATETIME_TYPE:
      case CS_DATETIME4_TYPE:
	len = 30;
	break;
	
      case CS_NUMERIC_TYPE:
      case CS_DECIMAL_TYPE:
	len = (CS_MAX_PREC + 2);
	break;
	
      default:
	len = column->maxlength;
	break;
    }
    
    return len;
}


static CS_RETCODE
describe(DBH *dbh, int restype)
{
    CS_RETCODE retcode;
    int i;
    int numCols;
    
    if((retcode = ct_res_info(dbh->cmd, CS_NUMDATA,
			      &numCols, CS_UNUSED, NULL)) != CS_SUCCEED)
    {
	warn("ct_res_info() failed");
	goto GoodBye;
    }
    if(numCols <= 0) {
	warn("ct_res_info() returned 0 columns");
	dbh->numCols = 0;
	goto GoodBye;
    }
    if(dbh->debug >= 2)
	fprintf(stderr, "    ct_res_info() returns %d columns\n", numCols);

    dbh->numCols = numCols;
    
    dbh->coldata = calloc(numCols, sizeof(ColData));
    dbh->datafmt = calloc(numCols, sizeof(CS_DATAFMT));
    
    /* this routine may be called without the connection reference */
    if(restype == CS_COMPUTE_RESULT) {
	CS_INT comp_id, outlen;
    
	if((retcode = ct_compute_info(dbh->cmd, CS_COMP_ID, CS_UNUSED,
			      &comp_id, CS_UNUSED, &outlen)) != CS_SUCCEED)
	{
	    warn("ct_compute_info failed");
	    goto GoodBye;
	}
    }

    for(i = 0; i < numCols; ++i) {
	if((retcode = ct_describe(dbh->cmd, (i + 1),
				  &dbh->datafmt[i])) != CS_SUCCEED)
	{
	    warn("ct_describe() failed");
	    clean_up(dbh);
	    goto GoodBye;
	}
	/* Make sure we have at least some sort of column name: */
	if(dbh->datafmt[i].namelen == 0)
	    sprintf(dbh->datafmt[i].name, "COL(%d)", i+1);
	if(restype == CS_COMPUTE_RESULT) {
	    CS_INT agg_op, outlen;
	    CS_CHAR *agg_op_name;
	    
	    if((retcode = ct_compute_info(dbh->cmd, CS_COMP_OP, (i + 1),
				  &agg_op, CS_UNUSED, &outlen)) != CS_SUCCEED)
	    {
		warn("ct_compute_info failed");
		goto GoodBye;
	    }
	    agg_op_name = get_agg_op(agg_op);
	    if((retcode = ct_compute_info(dbh->cmd, CS_COMP_COLID, (i + 1),
				  &agg_op, CS_UNUSED, &outlen)) != CS_SUCCEED)
	    {
		warn("ct_compute_info failed");
		goto GoodBye;
	    }
	    sprintf(dbh->datafmt[i].name, "%s(%d)", agg_op_name, 
		    (int)agg_op);
	}

	if(dbh->debug >= 3)
	    fprintf(stderr, "    ct_describe(%d): type = %d, maxlen = %d\n",
		    i, (int)dbh->datafmt[i].datatype,
		    (int)dbh->datafmt[i].maxlength);

	dbh->coldata[i].realType = dbh->datafmt[i].datatype;
	dbh->coldata[i].realLength = dbh->datafmt[i].maxlength;

	switch(dbh->datafmt[i].datatype) 
	{
	  case CS_BIT_TYPE:
	  case CS_TINYINT_TYPE:
	  case CS_SMALLINT_TYPE:
	  case CS_INT_TYPE:
	    dbh->datafmt[i].maxlength = sizeof(CS_INT);
	    dbh->datafmt[i].format    = CS_FMT_UNUSED;
	    dbh->coldata[i].type      = CS_INT_TYPE;
	    dbh->datafmt[i].datatype  = CS_INT_TYPE;
	    retcode = ct_bind(dbh->cmd, (i + 1), &dbh->datafmt[i],
			      &dbh->coldata[i].value.i,
			      &dbh->coldata[i].valuelen,
			      &dbh->coldata[i].indicator);
	    break;
	    
	  case CS_REAL_TYPE:
	  case CS_FLOAT_TYPE:
	  case CS_MONEY_TYPE:
	  case CS_MONEY4_TYPE:
	    dbh->datafmt[i].maxlength = sizeof(CS_FLOAT);
	    dbh->datafmt[i].format    = CS_FMT_UNUSED;
	    dbh->coldata[i].type      = CS_FLOAT_TYPE;
	    dbh->datafmt[i].datatype  = CS_FLOAT_TYPE;
	    retcode = ct_bind(dbh->cmd, (i + 1), &dbh->datafmt[i],
			      &dbh->coldata[i].value.f,
			      &dbh->coldata[i].valuelen,
			      &dbh->coldata[i].indicator);
	    break;
	    
	  case CS_IMAGE_TYPE:
	    dbh->coldata[i].value.c = malloc(dbh->datafmt[i].maxlength * 
					     sizeof(char));
	    dbh->datafmt[i].format   = CS_FMT_UNUSED; /*CS_FMT_NULLTERM;*/
	    dbh->coldata[i].type     = dbh->datafmt[i].datatype;
	    retcode = ct_bind(dbh->cmd, (i + 1), &dbh->datafmt[i],
			      dbh->coldata[i].value.c,
			      &dbh->coldata[i].valuelen,
			      &dbh->coldata[i].indicator);
	    break;
	    
	  case CS_CHAR_TYPE:
	  case CS_VARCHAR_TYPE:
	  case CS_BINARY_TYPE:
	  case CS_VARBINARY_TYPE:
	  case CS_NUMERIC_TYPE:
	  case CS_DECIMAL_TYPE:
	  case CS_DATETIME_TYPE:
	  case CS_DATETIME4_TYPE:
	  case CS_TEXT_TYPE:
	  default:
	    dbh->datafmt[i].maxlength = 
		get_cwidth(&dbh->datafmt[i]) + 1;
	    dbh->datafmt[i].format   = CS_FMT_NULLTERM;
	    dbh->coldata[i].value.c  = malloc(dbh->datafmt[i].maxlength * 
					      sizeof(char));
	    dbh->coldata[i].type      = CS_CHAR_TYPE;
	    dbh->datafmt[i].datatype  = CS_CHAR_TYPE;
	    retcode = ct_bind(dbh->cmd, (i + 1), &dbh->datafmt[i],
			      dbh->coldata[i].value.c,
			      &dbh->coldata[i].valuelen,
			      &dbh->coldata[i].indicator);
	    break;
	}	
	/* check the return code of the call to ct_bind in the
	   switch above: */
	if (retcode != CS_SUCCEED) {
	    warn("ct_bind() failed");
	    clean_up(dbh);
	    break;
	}
	if(dbh->debug >= 2)
	    fprintf(stderr, "    describe() -> col %d, type %d, realtype %d\n", 
		    i, (int)dbh->coldata[i].type, (int)dbh->coldata[i].realType);
	
    }
  GoodBye:;
    if(retcode == CS_SUCCEED) {
	dbh->done_desc = 1;
    }
    return retcode == CS_SUCCEED;
}


static int
next_result(DBH *dbh, RESULT *retdata)
{
    CS_COMMAND *cmd = dbh->cmd;
    CS_INT      restype;
    CS_RETCODE  retcode;
    int         failFlag = 0;

    while((retcode = ct_results(cmd, &restype)) == CS_SUCCEED) {
	if(dbh->debug >= 2)
	    fprintf(stderr, "    next_result() -> ct_results(%d) == %d\n",
		    (int)restype, (int)retcode);

	if(restype == CS_CMD_FAIL)
	    failFlag = 1;
	switch(restype) 
	{
	  case CS_ROW_RESULT:
	  case CS_PARAM_RESULT:
	  case CS_STATUS_RESULT:
	  case CS_CURSOR_RESULT:
	  case CS_COMPUTE_RESULT:
	    retcode = describe(dbh, restype);
	    if(dbh->debug >= 2)
		fprintf(stderr, "describe() retcode = %d\n", (int)retcode);
	    
	    if(restype == CS_STATUS_RESULT)
	    {
		CS_INT rows_read;
		retcode = ct_fetch(cmd, CS_UNUSED, CS_UNUSED, CS_UNUSED, &rows_read);
		if(retcode == CS_SUCCEED) {
		    retdata->procStatus = dbh->coldata[0].value.i;
		    retdata->hasProcStatus = 1;
		    if(dbh->debug >= 2)
			fprintf(stderr, "describe() proc status code = %d\n", retdata->procStatus);
		    if(retdata->procStatus != 0) {
			failFlag = 1;
		    }
		} else {
		    croak("ct_fetch() for proc status failed!");
		}
		while((retcode = ct_fetch(cmd, CS_UNUSED, CS_UNUSED, CS_UNUSED, &rows_read))) {
		    if(retcode == CS_END_DATA || retcode == CS_FAIL)
			break;
		}
	    } else {
		goto Done;	/* exit from the ct_results() loop here if 
				   this is *NOT* a status result set */
	    }
	}
    }
    if(dbh->debug >= 2)
	fprintf(stderr, "ct_execute() final retcode = %d\n", (int)retcode);
  Done:

    if((restype == CS_ROW_RESULT    ||
	restype == CS_PARAM_RESULT  ||
	restype == CS_CURSOR_RESULT ||
	restype == CS_COMPUTE_RESULT) && !failFlag ) {
	RESULT_LIST *res = calloc(1, sizeof(RESULT_LIST));

	if(dbh->debug >= 2)
	    fprintf(stderr, "restype = %d - adding RESULT to retdata\n",
		    (int)restype);

	res->resNum = ++retdata->numResults;
	res->numCols = dbh->numCols;

	if(!retdata->result) {
	    retdata->result = res;
	} else {
	    RESULT_LIST *tmp = retdata->result;
	    while(tmp->next) {
		tmp = tmp->next;
	    }
	    tmp->next = res;
	}
	    
	retdata->current = res;
    }

    if(failFlag || retcode == CS_FAIL)
	return CS_CMD_FAIL;

    dbh->lastResType = restype;
  
    return restype;
}



RESULT *db_exec_sql(DBH *dbh, char *sql) 
{
    RESULT *retdata;
    ROW *row;
    CS_RETCODE retcode;
    CS_INT rows_read, restype;
    int i;
    int fetching;

    if(dbh->debug>=2)
	fprintf(stderr, "db_exec_sql(%s)\n", sql);

    if(ct_command(dbh->cmd, CS_LANG_CMD, sql,
		  CS_NULLTERM, CS_UNUSED) != CS_SUCCEED) {
	return NULL;
    }

    if(ct_send(dbh->cmd) != CS_SUCCEED) {
	return NULL;
    }

    retdata = calloc(1, sizeof(RESULT));
    
    while((restype = next_result(dbh, retdata))) {
	if(restype == CS_CMD_DONE || restype == CS_CMD_FAIL) {
	    break;
	}
	fetching = 1;
	while(fetching) {
	    retcode = ct_fetch(dbh->cmd, CS_UNUSED, CS_UNUSED, CS_UNUSED,
			       &rows_read);
	    switch(retcode) {
	      case CS_ROW_FAIL:
		break;		// this happens if data is truncated.
	      case CS_SUCCEED:
		row = calloc(1, sizeof(ROW));
		row->data = calloc(retdata->current->numCols, 
				   sizeof(ColData));
		row->resNum = retdata->current->resNum;
		row->numCols = retdata->current->numCols;
		if(retdata->current->last != NULL) {
		    row->rowNum = retdata->current->last->rowNum + 1;
		    retdata->current->last->next = row;
		    retdata->current->last = row;
		} else {
		    if(retdata->current->row) {
			row->rowNum = 2;
			retdata->current->row->next = row;
			retdata->current->last = row;
		    } else {
			row->rowNum = 1;
			retdata->current->row = row;
		    }
		}
		    
		for(i = 0; i < row->numCols; ++i) {
		    row->data[i] = dbh->coldata[i];
		    if(row->data[i].type == CS_CHAR_TYPE) {
			row->data[i].value.c = strdup(dbh->coldata[i].value.c);
		    }
		}
		break;
	      case CS_FAIL:		/* ohmygod */
		/* FIXME: Should we call ct_cancel() here, or should we let
		   the programmer handle it? */
		if(ct_cancel(dbh->connection, NULL, CS_CANCEL_ALL) == CS_FAIL) {
		    ct_close(dbh->connection, CS_FORCE_CLOSE);
		    dbh->isDead = 1;
		}
		return NULL;
		break;
	      case CS_END_DATA:	/* we've seen all the data for this result
				   set. So see if this is the end of the
				   result sets */
		fetching = 0;	// break out of fetch loop.
		break;
	      default:
		warn("ct_fetch() returned an unexpected retcode");
	    }
	}
    }

    return retdata;
}

void db_free_result(RESULT *retdata) {
    RESULT_LIST *result = retdata->result;
    ROW *row;
    int i;
    void *next;

    while(result) {
	row = result->row;
	while(row) {
	    for(i = 0; i < row->numCols; ++i) {
		if(row->data[i].type == CS_CHAR_TYPE ||
		   row->data[i].type == CS_IMAGE_TYPE)
		    free(row->data[i].value.c);
	    }
	    next = row->next;
	    free(row->data);
	    free(row);
	    row = next;
	}
	next = result->next;
	free(result);
	result = next;
    }
    free(retdata);
}

void db_use (DBH *dbh, char *database) 
{
    char sql[60];
    RESULT *ret;
    
    sprintf(sql, "use %s", database);
    ret = db_exec_sql(dbh, sql);
    db_free_result(ret);
}

void db_free (DBH *dbh)
{
    ct_cmd_drop(dbh->cmd);
    ct_close(dbh->connection, CS_FORCE_CLOSE);
    ct_con_drop(dbh->connection);
    cs_loc_drop(context, dbh->locale);

    free(dbh);
}

void db_dump(RESULT *retdata) {
    RESULT_LIST *result = retdata->result;
    ROW *row;
    int i;

    while(result) {
	printf("Dumping result set %d\n", result->resNum);
	row = result->row;
	while(row) {
	    printf("%5d", row->rowNum);
	    for(i = 0; i < row->numCols; ++i) {
		switch(row->data[i].type) {
		  case CS_CHAR_TYPE:
		    printf(" %s", row->data[i].value.c);
		    break;
		  case CS_INT_TYPE:
		    printf(" %d", (int)row->data[i].value.i);
		    break;
		  case CS_FLOAT_TYPE:
		    printf(" %f", row->data[i].value.f);
		    break;
		}
	    }
	    printf("\n");
	    row = row->next;
	}

	result = result->next;
    }
    if(retdata->hasProcStatus) {
	printf("Return status: %d\n", retdata->procStatus);
    }
    
}
    
    
